/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <platform.h>

#include <algorithm>
#include <fstream>

#include <mapi.h>
#include <mapiext.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapidefs.h>
#include <mapiguid.h>

#include <edkguid.h>
#include <edkmdb.h>

#include <ECGuid.h>
#include <ECSerializer.h>
#include <ECTags.h>
#include <IECMultiStoreTable.h>
#include <stringutil.h>
#include <Util.h>

#include "ECFileIndex.h"
#include "ECIndexerUtil.h"
#include "ECLucene.h"
#include "ECLuceneAccess.h"
#include "ECLuceneIndexer.h"
#include "ECLuceneIndexerAttachment.h"
#include "zarafa-indexer.h"

ECLuceneIndexer::ECLuceneIndexer(ECThreadData *lpThreadData, ECLuceneAccess *lpLuceneAccess)
{
	m_lpThreadData = lpThreadData;

	m_lpLuceneAccess = lpLuceneAccess;
	m_lpLuceneAccess->AddRef();
}

ECLuceneIndexer::~ECLuceneIndexer()
{
	if (m_lpIndexedProps)
		MAPIFreeBuffer(m_lpIndexedProps);

	if (m_lpLuceneAccess)
		m_lpLuceneAccess->Release();

	if (m_lpIndexerAttach)
		m_lpIndexerAttach->Release();

	if (m_lpMsgStore)
		m_lpMsgStore->Release();

	if (m_lpEMS)
		m_lpEMS->Release();
}

HRESULT ECLuceneIndexer::Create(ECThreadData *lpThreadData, ECLuceneAccess *lpLuceneAccess, IMsgStore *lpMsgStore, ECLuceneIndexer **lppIndexer)
{
	HRESULT hr = hrSuccess;
	ECLuceneIndexer *lpIndexer = NULL;

	SizedSPropTagArray(2, sExtra) = {
		2, {
			PR_HASATTACH,
			PR_MESSAGE_CLASS_A,
		},
	};

	lpIndexer = new ECLuceneIndexer(lpThreadData, lpLuceneAccess);
	if (!lpIndexer) {
		hr = MAPI_E_NOT_ENOUGH_MEMORY;
		goto exit;
	}

	lpIndexer->AddRef();

	hr = lpIndexer->m_lpThreadData->lpLucene->GetIndexedProps((LPSPropTagArray)&sExtra, &lpIndexer->m_lpIndexedProps);
	if (hr != hrSuccess)
		goto exit;

	hr = ECLuceneIndexerAttachment::Create(lpIndexer->m_lpThreadData, lpIndexer, &lpIndexer->m_lpIndexerAttach);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMsgStore->QueryInterface(IID_IMsgStore, (LPVOID *)&lpIndexer->m_lpMsgStore);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMsgStore->QueryInterface(IID_IExchangeManageStore, (LPVOID*)&lpIndexer->m_lpEMS);
	if (hr != hrSuccess)
		goto exit;

	if (lppIndexer)
		*lppIndexer = lpIndexer;

exit:
	if ((hr != hrSuccess) && lpIndexer)
		lpIndexer->Release();

	return hr;
}

HRESULT ECLuceneIndexer::IndexEntries(sourceid_list_t &listCreateEntries, sourceid_list_t &listChangeEntries, sourceid_list_t &listDeleteEntries)
{
	HRESULT hr = hrSuccess;

	try {
		if (m_lpLuceneAccess->IndexExists()) {
			if (!listDeleteEntries.empty()) {
				hr = IndexDeleteEntries(listDeleteEntries);
				if (hr != hrSuccess)
					goto exit;
			}

			/* We can't dynamically update an index message, it must be deleted and recreated */
			if (!listChangeEntries.empty()) {
				hr = IndexDeleteEntries(listChangeEntries);
				if (hr != hrSuccess)
					goto exit;
			}
		}

		if (!listCreateEntries.empty()) {
			hr = IndexCreateEntries(listCreateEntries);
			if (hr != hrSuccess)
				goto exit;

			hr = IndexStreamEntries(listCreateEntries);
			if (hr != hrSuccess)
				goto exit;
		}

		/* We can't dynamically update an index message, it must be deleted and recreated */
		if (!listChangeEntries.empty()) {
			hr = IndexCreateEntries(listChangeEntries);
			if (hr != hrSuccess)
				goto exit;

			hr = IndexStreamEntries(listChangeEntries);
			if (hr != hrSuccess)
				goto exit;

		}
	} catch (CLuceneError &e) {
		m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_FATAL, "CLucene error: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	return hr;
}


HRESULT ECLuceneIndexer::IndexDeleteEntries(sourceid_list_t &listSourceId)
{
	HRESULT hr = hrSuccess;
	lucene::index::IndexReader *lpReader = NULL;
	lucene::index::Term sTerm;
	sourceid_list_t::iterator iter;
	NShttpmail_t * lpNSProp = NULL;
	unsigned int deleted = 0;

	lpNSProp = m_lpThreadData->lpLucene->GetIndexedProp(PR_SOURCE_KEY);
	if (!lpNSProp)
		goto exit;

	hr = m_lpLuceneAccess->GetLuceneReader(&lpReader);
	if (hr != hrSuccess)
		goto exit;

	for (iter = listSourceId.begin(); iter != listSourceId.end(); iter++) {
		std::wstring strField = lpNSProp->strNShttpmail;
		std::wstring strContents = PropToString(&(*iter)->lpProps[0]);

		sTerm.set(strField.c_str(), strContents.c_str());

		try {
			unsigned int d = lpReader->deleteDocuments(&sTerm);
			deleted += d;
		}
		catch (CLuceneError &e) {
			m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_FATAL, "CLucene error: %s", e.what());
			continue;
		}
	}
	
exit:
	if (lpReader)
		m_lpLuceneAccess->PutLuceneReader(lpReader, deleted > 0);

	return hr;
}

HRESULT ECLuceneIndexer::IndexCreateEntries(sourceid_list_t &listSourceId)
{
	HRESULT hr = hrSuccess;
	lucene::index::IndexWriter *lpWriter = NULL;
	SBinaryArray *lpMSTContent = NULL;
	IECMultiStoreTable *lpMSTable = NULL;
	IMAPITable *lpTable = NULL;
	LPSRowSet lpRows = NULL;

	hr = m_lpLuceneAccess->GetLuceneWriter(&lpWriter);
	if (hr != hrSuccess)
		goto exit;

	hr = CreateMultiStoreTableContents(listSourceId, &lpMSTContent);
	if (hr != hrSuccess)
		goto exit;

	/* Nothing to do */
	if (lpMSTContent->cValues == 0)
		goto exit;

	hr = m_lpMsgStore->QueryInterface(IID_IECMultiStoreTable, (LPVOID *)&lpMSTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMSTable->OpenMultiStoreTable(lpMSTContent, EC_TABLE_NOCAP, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns(m_lpIndexedProps, TBL_BATCH);
	if (hr != hrSuccess)
		goto exit;

	while (TRUE) {
		hr = lpTable->QueryRows(250, 0, &lpRows);
		if (hr != hrSuccess)
			goto exit;

		if (lpRows->cRows == 0)
			break;

		for (ULONG i = 0; i < lpRows->cRows; i++) {
			lucene::document::Document sDocument;

			hr = ParseDocument(&sDocument, lpRows->aRow[i].cValues, lpRows->aRow[i].lpProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			try {
				lpWriter->addDocument(&sDocument);
			}
			catch (CLuceneError &e) {
				m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_FATAL, "CLucene error: %s", e.what());
				goto exit;
			}
		}

		if (lpRows)
			FreeProws(lpRows);
		lpRows = NULL;
	}

exit:
	if (lpWriter)
		m_lpLuceneAccess->PutLuceneWriter(lpWriter);

	if (lpRows)
		FreeProws(lpRows);

	if (lpMSTable)
		lpMSTable->Release();

	if (lpTable)
		lpTable->Release();

	if (lpMSTContent)
		DeleteMultiStoreTableContents(lpMSTContent);

	return hr;
}

HRESULT ECLuceneIndexer::IndexStreamEntries(sourceid_list_t &listSourceId)
{
	HRESULT hr = hrSuccess;
	ECSerializer *lpSerializer = NULL;
	lucene::index::IndexWriter *lpWriter = NULL;
	LARGE_INTEGER lint = {{ 0, 0 }};

	hr = m_lpLuceneAccess->GetLuceneWriter(&lpWriter);
	if (hr != hrSuccess)
		goto exit;

	lpSerializer = new ECStreamSerializer(NULL);
	if (!lpSerializer) {
		hr = MAPI_E_NOT_ENOUGH_MEMORY;
		goto exit;
	}

	for (sourceid_list_t::iterator iter = listSourceId.begin(); iter != listSourceId.end(); iter++) {
		lucene::document::Document sDocument;

		/* No stream, will be handled by IndexCreateEntries */
		if (!(*iter)->lpStream)
			continue;

		/* Read number of properties from Stream */
		hr = (*iter)->lpStream->Seek(lint, STREAM_SEEK_SET, NULL);
		if (hr != hrSuccess)
			goto exit;

		lpSerializer->SetBuffer((*iter)->lpStream);

		hr = ParseStream(&sDocument, (*iter)->ulProps, (*iter)->lpProps, lpSerializer);
		if (hr != hrSuccess)
			goto exit;

		try {
			lpWriter->addDocument(&sDocument);
		}
		catch (CLuceneError &e) {
			m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_FATAL, "CLucene error: %s", e.what());
			goto exit;
		}
	}

exit:
	if (lpSerializer)
		delete lpSerializer;

	if (lpWriter)
		m_lpLuceneAccess->PutLuceneWriter(lpWriter);

	return hr;
}

HRESULT ECLuceneIndexer::ParseDocument(lucene::document::Document *lpDocument, ULONG cValues, LPSPropValue lpProps, IMessage *lpMessage, BOOL bDefaultField)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpEntryId = NULL;
	LPSPropValue lpMsgClass = NULL;
	BOOL bAttach = FALSE;
	
	lpMsgClass = PpropFindProp(lpProps, cValues, PR_MESSAGE_CLASS_A);
	if (lpMsgClass && strcmp(lpMsgClass->Value.lpszA, "IPM.Zarafa.Stub") == 0) {
		m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_ANY, "Ignoring stubbed message.");
		goto exit;
	}

	for (ULONG i = 0; i < cValues; i++) {
		/* PR_HASATTACH should not be indexed */
		if (lpProps[i].ulPropTag == PR_HASATTACH) {
			bAttach = lpProps[i].Value.b;
			continue;
		}

		/* We might need the entryid later */
		if (lpProps[i].ulPropTag == PR_ENTRYID)
			lpEntryId = &lpProps[i];

		/* Index property */
		hr = AddPropertyToDocument(lpDocument, &lpProps[i], bDefaultField);
		if (hr != hrSuccess)
			goto exit;
	}

	if (bAttach && (lpEntryId || lpMessage)) {
		hr = ParseAttachments(lpDocument, lpEntryId, lpMessage);
		if (hr != hrSuccess) {
			/* Ignore error, just index message without attachments */
			hr = hrSuccess;
		}
	}

exit:
	return hr;
}

HRESULT ECLuceneIndexer::ParseStream(lucene::document::Document *lpDocument, ULONG cValues, LPSPropValue lpProps, ECSerializer *lpSerializer, BOOL bDefaultField)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpProp = NULL;
	LPSPropValue lpMsgClass = NULL;
	ULONG ulStreamVersion = 0;
	ULONG ulProps = 0;
	
	lpMsgClass = PpropFindProp(lpProps, cValues, PR_MESSAGE_CLASS_A);
	if (lpMsgClass && strcmp(lpMsgClass->Value.lpszA, "IPM.Zarafa.Stub") == 0) {
		m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_ANY, "Ignoring stubbed message.");
		FlushStream(lpSerializer);
		m_lpThreadData->lpLogger->Log(EC_LOGLEVEL_DEBUG, "Flushed stream.");
		goto exit;
	}

	for (unsigned int i = 0; i < cValues; i++) {
		/* Index property */
		hr = AddPropertyToDocument(lpDocument, &lpProps[i], bDefaultField);
		if (hr != hrSuccess)
			goto exit;
	}

	/* Only the toplevel contains the stream version */
	if (!bDefaultField) {
		if (lpSerializer->Read(&ulStreamVersion, sizeof(ulStreamVersion), 1) != erSuccess) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		if (ulStreamVersion != 0) {
			hr = MAPI_E_NO_SUPPORT;
			goto exit;
		}
	}

	if (lpSerializer->Read(&ulProps, sizeof(ulProps), 1) != erSuccess) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	for (unsigned int i = 0; i < ulProps; i++) {
		/* Make sure we start with empty variables */
		if (lpProp)
			MAPIFreeBuffer(lpProp);
		lpProp = NULL;

		/* Read property from stream */
		hr = StreamToProperty(lpSerializer, &lpProp);
		if (hr != hrSuccess) {
			if (hr == MAPI_E_NO_SUPPORT) {
				hr = hrSuccess;
				continue;
			}
			goto exit;
		}

		/*
		 * Stream syncer sends either PR_BODY, PR_HTML, PR_BODY_HTML or PR_RTF_COMPRESSED,
		 * if multiple are available, only the "best" property will be provided. Since
		 * we only want PR_BODY we need to perform conversions in case the HTML or RTF
		 * properties were provided instead
		 */
		if ((lpProp->ulPropTag == PR_HTML) ||
			(lpProp->ulPropTag == PR_BODY_HTML) ||
			(lpProp->ulPropTag == PR_RTF_COMPRESSED))
				ConvertToPlainText(lpProp);
		
		/* Index property */
		hr = AddPropertyToDocument(lpDocument, lpProp, bDefaultField);
		if (hr != hrSuccess)
			goto exit;
	}

	hr = ParseStreamAttachments(lpDocument, lpSerializer);
	if (hr != hrSuccess) {
		/* Ignore error, just index message without attachments */
		hr = hrSuccess;
	}

exit:
	if (lpProp)
		MAPIFreeBuffer(lpProp);

	return hr;
}

HRESULT ECLuceneIndexer::ParseAttachments(lucene::document::Document *lpDocument, LPSPropValue lpEntryId, IMessage *lpOrigMessage)
{
	HRESULT hr = hrSuccess;
	IMessage *lpMessage = NULL;
	ULONG ulObjType = 0;

	if (!parseBool(m_lpThreadData->lpConfig->GetSetting("index_attachments"))) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	if (lpOrigMessage) {
		lpMessage = lpOrigMessage;
		lpMessage->AddRef();
	} else if (lpEntryId && lpEntryId->Value.bin.cb) {
		hr = m_lpMsgStore->OpenEntry(lpEntryId->Value.bin.cb, (LPENTRYID)lpEntryId->Value.bin.lpb, &IID_IMessage, 0, &ulObjType, (LPUNKNOWN *)&lpMessage);
		if (hr != hrSuccess)
			goto exit;
	} else {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	hr = m_lpIndexerAttach->ParseAttachments(lpDocument, lpMessage);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (lpMessage)
		lpMessage->Release();

	return hr;
}

HRESULT ECLuceneIndexer::ParseStreamAttachments(lucene::document::Document *lpDocument, ECSerializer *lpSerializer)
{
	HRESULT hr = hrSuccess;

	if (!parseBool(m_lpThreadData->lpConfig->GetSetting("index_attachments"))) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	hr = m_lpIndexerAttach->ParseAttachments(lpDocument, lpSerializer);
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}	

HRESULT ECLuceneIndexer::AddPropertyToDocument(lucene::document::Document *lpDocument, LPSPropValue lpProp, BOOL bDefaultField)
{
	HRESULT hr = hrSuccess;
	lucene::document::Field *lpField = NULL;
	NShttpmail_t * lpNSProp = NULL;
	std::wstring strField;
	std::wstring strContents;

	/* Check if this property needs to be indexed */
	lpNSProp = m_lpThreadData->lpLucene->GetIndexedProp(lpProp->ulPropTag);
	if (!lpNSProp)
		goto exit;

	strField = bDefaultField ? DEFAULT_FIELD : lpNSProp->strNShttpmail;
	if (strField.empty())
		goto exit;

	strContents = PropToString(lpProp);
	if (strContents.empty())
		goto exit;

	try {
		lpField = new lucene::document::Field(strField.c_str(), strContents.c_str(), lpNSProp->ulStorage);
	}
	catch (...) {
		hr = MAPI_E_NOT_ENOUGH_MEMORY;
		goto exit;
	}
	if (lpField)
		lpDocument->add(*lpField);

exit:
	return hr;
}

HRESULT ECLuceneIndexer::CreateMultiStoreTableContents(sourceid_list_t &listSourceId, SBinaryArray **lppMSTContent)
{
	HRESULT hr = hrSuccess;
	SBinaryArray *lpMSTContent = NULL;
	LPSPropValue lpEntryId = NULL;
	LPSPropValue lpSourceKey = NULL;
	LPSPropValue lpParentSourceKey = NULL;

	hr = MAPIAllocateBuffer(sizeof(SBinaryArray), (LPVOID *)&lpMSTContent);
	if (hr != hrSuccess)
		goto exit;

	lpMSTContent->cValues = 0;

	hr = MAPIAllocateMore(sizeof(SBinary) * listSourceId.size(), lpMSTContent, (LPVOID *)&lpMSTContent->lpbin);
	if (hr != hrSuccess)
		goto exit;

	for (sourceid_list_t::iterator iter = listSourceId.begin(); iter != listSourceId.end(); iter++) {
		/* Stream has preference over OpenEntry() */
		if ((*iter)->lpStream)
			continue;

		lpEntryId = PpropFindProp((*iter)->lpProps, (*iter)->ulProps, PR_ENTRYID);
		if (lpEntryId) {
			/* Entryid is known, copy entryid */
			hr = Util::HrCopyBinary(lpEntryId->Value.bin.cb, lpEntryId->Value.bin.lpb,
									&lpMSTContent->lpbin[lpMSTContent->cValues].cb, &lpMSTContent->lpbin[lpMSTContent->cValues].lpb);
			if (hr != hrSuccess)
				goto exit;
		} else {
			/* If entryid is not known, we must use the Sourcekey */
			lpSourceKey = PpropFindProp((*iter)->lpProps, (*iter)->ulProps, PR_SOURCE_KEY);
			lpParentSourceKey = PpropFindProp((*iter)->lpProps, (*iter)->ulProps, PR_PARENT_SOURCE_KEY);

			/* No sourcekey information, this entry is invalid */
			if (!lpSourceKey || !lpParentSourceKey)
				continue;

			hr = m_lpEMS->EntryIDFromSourceKey(lpParentSourceKey->Value.bin.cb, lpParentSourceKey->Value.bin.lpb,
											   lpSourceKey->Value.bin.cb, lpSourceKey->Value.bin.lpb,
											   &lpMSTContent->lpbin[lpMSTContent->cValues].cb, (LPENTRYID *)&lpMSTContent->lpbin[lpMSTContent->cValues].lpb);
			if (hr != hrSuccess) {
				if (hr != MAPI_E_NOT_FOUND && hr != MAPI_E_INVALID_ENTRYID)
					goto exit;

				/* Mail was deleted, saves us some time indexing */
				hr = hrSuccess;
				continue;
			}
		}

		lpMSTContent->cValues++;
	}

	if (lppMSTContent)
		*lppMSTContent = lpMSTContent;

exit:
	if ((hr != hrSuccess) && lpMSTContent)
		DeleteMultiStoreTableContents(lpMSTContent);

	return hr;
}

VOID ECLuceneIndexer::DeleteMultiStoreTableContents(SBinaryArray *lpMSTContent)
{
	if (!lpMSTContent)
		return;

	for (ULONG i = 0; i < lpMSTContent->cValues; i++) {
		if (lpMSTContent->lpbin[i].lpb)
			MAPIFreeBuffer(lpMSTContent->lpbin[i].lpb);
	}

	MAPIFreeBuffer(lpMSTContent);
}
